from __future__ import print_function, division, absolute_import, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import struct
import numpy

# Contains code for unpacking data from SDM binary tables.


def unpacker(sdmtable):
    """Return an appropriate unpacker given an SDMBinaryTable object."""
    # Could do something more fancy with a registry, etc but it's probably
    # not worth it given the small number of binary tables.
    if sdmtable.name == 'SysPower':
        return SysPowerUnpacker(sdmtable)
    else:
        raise RuntimeError("Unknown binary table type: '%s'" % sdmtable.name)


class BinaryTableUnpacker(object):
    """Base class for SDM binary table unpackers.  This should not be called
    directly, because it needs a column data structure definition in order
    to be useful.  This is implemented in derived classes, e.g.
    SysPowerUnpacker for the SysPower table."""

    # Unpackers for various binary data types
    _unpack_int = struct.Struct('>i')
    _unpack_long = struct.Struct('>q')
    _unpack_float = struct.Struct('>f')

    # Column definitions go in derived classes, see SysPowerUnpacker
    # below for an example.
    columns = []

    def __init__(self, sdmtable):
        self._sdmtable = sdmtable  # The original SDMBinaryTable object
        self._pos = 0  # Pointer to current position within the table
        # Read the entity stuff at the start of the table, then
        # record the start of row data.  Just use tuples for the
        # entity things now, until building a better entity data
        # structure seems worthwhile.
        self.table_entity = tuple(self._get_val('S') for i in range(5))
        self.container_entity = tuple(self._get_val('S') for i in range(5))
        self.nrows = self._get_val('i4')  # -1 in all data I've looked at...
        self._pos0 = self._pos
        self.row = []

    def _readtab(self, nbytes):
        # Read/return nbytes from the table, and increment the
        # position marker.
        result = self._sdmtable.get_bytes(self._pos, nbytes)
        if len(result) != nbytes:
            return None
        self._pos += nbytes
        return result

    def _get_val(self, dtype, optional=False, array=False):
        # Unpack one value of the specified type from the table,
        # and advance the position pointer appropriately.  dtype
        # uses values compatible with numpy, ie i4, f4, etc.
        string_val = False
        if dtype == 'i4':
            unpack = self._unpack_int
        elif dtype == 'i8':
            unpack = self._unpack_long
        elif dtype == 'f4':
            unpack = self._unpack_float
        elif dtype[0] == 'S':
            unpack = None
            string_val = True
        else:
            raise RuntimeError("Unknown data dtype (dtype='%s')" % dtype)
        try:
            if optional:
                has_data = self._readtab(1) != '\0'
                if not has_data:
                    # Return some default values..
                    if string_val:
                        return ''
                    else:
                        return 0
            if array or string_val:
                nval = self._unpack_int.unpack_from(self._readtab(4))[0]
            if string_val:
                return self._readtab(nval)
            dsize = unpack.size
            if array:
                return [unpack.unpack_from(self._readtab(dsize))[0]
                        for i in range(nval)]
            else:
                return unpack.unpack_from(self._readtab(dsize))[0]
        except struct.error:
            # Not enough bytes to unpack probably means we are at the end of
            # the table.
            return None

    @property
    def record_dtype(self):
        return [(col[0], col[2], col[3]) for col in self.columns]

    def _blank_row(self, nrows=1):
        return numpy.zeros(nrows, dtype=self.record_dtype)

    def _unpack_row(self):
        # Read the next row starting at current position and fill
        # into a single-row record array.
        row = self._blank_row()
        for col in self.columns:
            # Maybe use a better data struct for these...
            colname = col[0]
            isoptional = col[1]
            coldtype = col[2]
            isarray = col[3] != ()
            val = self._get_val(coldtype, isoptional, isarray)
            if val is None:
                return None
            if isarray:
                row[0][colname][:len(val)] = val
            else:
                row[0][colname] = val
        return row

    def iterrows(self):
        # Iterator over rows.  May or may not make sense to expose this
        # since everything is probably faster on the full unpacked array..
        self._pos = self._pos0  # reset to start
        more_data = True
        while more_data:
            row = self._unpack_row()
            if row is None:
                more_data = False
            else:
                yield row

    def unpack(self):
        self.row = numpy.recarray(1000, dtype=self.record_dtype)
        irow = 0
        for row in self.iterrows():
            if irow == len(self.row):
                self.row = numpy.rec.array(
                        numpy.resize(self.row, int(len(self.row)*1.2))
                        )
            self.row[irow] = row
            irow += 1
        self.row = numpy.rec.array(numpy.resize(self.row, irow))


class SysPowerUnpacker(BinaryTableUnpacker):
    """Unpacker for the SysPower binary table.  Currently makes some
    assumptions like there never being more than 2 measurements in
    the array-valued columns.  This is true for now for the VLA.
    """

    # Define columns using tuples like:
    #   (colname, optional, dtype, array_dims)
    # The column names do not have to exactly match the SDM definition
    # but seems like a reasonable idea to stay close.  The data types
    # should be recognized by numpy.recarray dtype argument.  This means
    # we need to specify an array dimension here.  array_dims should be set
    # to () for scalar columns.

    columns = [
            ("antennaId",        False, 'S32', ()),
            ("spectralWindowId", False, 'S32', ()),
            ("feedId",           False, 'i4', ()),
            ("timeMid",          False, 'i8', ()),
            ("interval",         False, 'i8', ()),
            ("numReceptor",      False, 'i4', ()),
            ("switchedPowerDifference", True, 'f4', (2,)),
            ("switchedPowerSum",        True, 'f4', (2,)),
            ("requantizerGain",         True, 'f4', (2,)),
            ]



#! /usr/bin/env python

# Split a phase-binned SDM/BDF set into multiple single-bin
# sets.  Currently requires all scans/spws in the set to use the same
# number of bins.
# PBD 06/2016

import os, sys
from collections import namedtuple
import numpy as np
from numpy.fft import fft, ifft
np.errstate(divide='ignore')
import sdmpy
from sdmpy.pulsar import dedisperse_array
import progressbar
import argparse

par = argparse.ArgumentParser()
par.add_argument("sdmname", help="SDM to process")
par.add_argument("-d", "--dm", type=float, default=0.0,
        help="dispersion measure (pc cm^-3) [%(default)s]")
par.add_argument("-p", "--period", type=float, default=1.0,
        help="pulse period (s) [%(default)s]")
par.add_argument("-m", "--meansub", action="store_true",
        help="subtract period-averaged value from each bin")
par.add_argument("-s", "--scan", action="append", default=[],
        help="process this scan (multiple -s allowed)")
par.add_argument("-C", "--cal", action="store_true",
        help="produce sum/diff outputs (cal mode)")
par.add_argument("-A", "--fixautos", action="store_true",
        help="attempt to fix sign of autocorr poln products")
par.add_argument("-T", "--template", type=str, default='',
        help="convolve with ascii template given in filename")
args = par.parse_args()

sdmname = args.sdmname.rstrip('/')

sdm = sdmpy.SDM(sdmname)

# Get the number of bins, currently assumes it's constant throughout
# the data set.
# TODO: when processing a multi-bin data set this will need to change,
# I guess the right thing to do is get the maximum number of bins
# during the observation and set up enough output files to handle
# that case.
try:
    bdf0 = sdm.scan(args.scan[0]).bdf
except IndexError:
    bdf0 = sdm.scan(1).bdf
nbin = bdf0.spws[0].numBin

# Read a template file
tmpl = None
if args.template:
    # Assume this is a simple ascii file of template values
    tmpl = np.loadtxt(args.template,usecols=0)
    # Check that number of bins matches.  Could implement template rebinning
    # at some point.
    if len(tmpl) != nbin:
        print "Error, number of template bins (%d) does not match data (%d)" % (
                len(tmpl), nbin)
        sys.exit(1)
    # Rescale template values to preserve mean flux, and mean-subtract:
    tmpl -= tmpl.min()
    tmpl /= tmpl.sum()
    tmpl -= tmpl.mean()
    # FFT to prepare for convolution later on.  reshape so that it 
    # will broadcast correctly against a full data array.
    ftmpl = np.conj(fft(tmpl)).reshape((1,1,nbin,1,1))

# Sets up the output paths.  Use the "extra" ibin==nbin entry
# for the averaged data.
sdmout = []
bdfoutpath = []
if args.cal:
    sdmout.append(sdmname + '.sum')
    sdmout.append(sdmname + '.dif')
    nbin = 2
else:
    for ibin in range(nbin+1):
        if ibin==nbin:
            sdmout.append(sdmname + '.avg')
        else:
            sdmout.append(sdmname + '.bin%04d'%ibin)

for ibin in range(len(sdmout)):
    bdfoutpath.append(sdmout[ibin] + '/ASDMBinary')
    os.mkdir(sdmout[ibin])
    os.mkdir(bdfoutpath[ibin])

for scan in sdm.scans():
    if len(args.scan) and scan.idx not in args.scan:
        continue
    print "Processing '%s' scan %s:" % (sdmname, scan.idx)
    try:
        bdf = scan.bdf
    except IOError:
        print "Error reading bdf for scan %s, skipping" % (scan.idx,)
        continue
    if args.cal and bdf.spws[0].numBin != 2:
        print "nbin!=2 for scan %s, skipping" % (scan.idx,)
        continue
    # Array of dims (nspw,nchan) giving freqs in MHz
    freqs_spw = scan.freqs()/1e6
    bdfoutname = map(lambda x: x+'/'+os.path.basename(scan.bdf_fname), 
            bdfoutpath)
    # Set up for output BDFs, copying header info from the input BDF
    # and changing nbin to 1.
    bdfout = map(lambda x: sdmpy.bdf.BDFWriter(x,bdf=bdf), bdfoutname)
    for ibdf in bdfout: 
        for bb in ibdf.sdmDataHeader.dataStruct.baseband:
            for spw in bb.spectralWindow:
                spw.attrib['numBin'] = '1'
        # update size attributes
        for a in ('flags','actualTimes','actualDurations',
                'crossData','autoData'):
            try:
                ds = ibdf.sdmDataHeader.dataStruct.__dict__[a]
                if 'BIN' in ds.attrib['axes']:
                    sz = int(ds.attrib['size'])
                    ds.attrib['size'] = str(sz/nbin)
            except KeyError:
                pass
        ibdf.write_header()
    bar = progressbar.ProgressBar()
    for i in bar(range(bdf.numIntegration)):
        fullint = bdf[i]
        dtypes = fullint.data.keys()
        binint = []
        for ibin in range(len(bdfout)):
            # This is kind of a kludge to create a new BDFInt-like container
            # to hold the data that will be written out.  May want to clean
            # this up at some point.
            binint.append(namedtuple("NewBDFIntegration",
                ['sdmDataSubsetHeader','data'])(
                    fullint.sdmDataSubsetHeader,
                    {}
                    )
                )
            for dtype in dtypes:
                binint[-1].data[dtype] = None
        for dtype in dtypes:

            ## This version handles the general case where different
            ## spws can potentially have different numbers of channels
            ## or polarizations, etc.:
            #for ispw in range(len(fullint.spws)):
            #    # Axes should be bl/ant, bin, chan, poln
            #    data = fullint.get_data(ispw,type=dtype)
            #    for ibin in range(nbin):
            #        bindata = data.take(ibin,
            #                axis=1).reshape((data.shape[0],-1))
            #        if binint[ibin].data[dtype] is None:
            #            binint[ibin].data[dtype] = bindata.copy()
            #        else:
            #            binint[ibin].data[dtype] = np.hstack(
            #                    (binint[ibin].data[dtype], bindata) 
            #                    )

            ## This much simpler version handles the case where all spws
            ## have matching dimensions.  Will raise an error if this 
            ## is not the case.
            # Axes should be (bl/ant, spw, bin, chan, pol)
            data = fullint.get_data(type=dtype).copy()

            # extend zeros along bin axis
            dataspw = data.sum(3,keepdims=True)
            dataz = dataspw!=0.0
            dataz_m = dataspw.all(2,keepdims=True)
            data *= dataz_m
            
            # Dedisperse if needed
            if args.dm!=0.0:
                dedisperse_array(data, args.dm, freqs_spw, args.period,
                        bin_axis=2, freq_axis=3, spw_axis=1)

            # Flip sign of autocorr cross-pol imag component in 
            # every other spw.  Note I don't think there is a 
            # general way to know which spws need fixing..
            if args.fixautos:
                if ('auto' in dtype) and (data.shape[-1]==4):
                    for ispw in range(1,data.shape[1],2):
                        data[:,ispw,:,:,2] *= -1.0

            if args.cal:
                # If either one of the bins is zeroed, zero both in
                # output.
                b0 = data.take(0,axis=2)
                b1 = data.take(1,axis=2)
                wt = (b0!=0.0) * (b1!=0.0)
                binint[0].data[dtype] = (wt*(b1+b0))*0.5
                binint[1].data[dtype] = wt*(b1-b0)
            else:
                binint[nbin].data[dtype] = data.mean(axis=2)
                if tmpl is not None:
                    cdata = ifft(fft(data,axis=2)*ftmpl,axis=2)
                    for ibin in range(nbin):
                        binint[ibin].data[dtype] = \
                                cdata.take(ibin,axis=2).astype(data.dtype)
                else:
                    for ibin in range(nbin):
                        binint[ibin].data[dtype] = data.take(ibin,axis=2)
                        if args.meansub:
                            binint[ibin].data[dtype] -= binint[nbin].data[dtype]

        for ibin in range(len(bdfout)):
            bdfout[ibin].write_integration(binint[ibin])
                
    for b in bdfout: b.close()

    # update SDM entries with corect data size (should be same for all bins)
    scan._main.dataSize = os.path.getsize(bdfoutname[0])

# Write out new SDM tables
for s in sdmout: sdm.write(s)




from __future__ import print_function, division, absolute_import, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import os
import sys
import re
import mmap
import math
import numpy
from copy import deepcopy
from lxml import etree, objectify

try:
    from progressbar import ProgressBar
except ImportError:
    ProgressBar = None

from .mime import MIMEPart, MIMEHeader

import logging
logger = logging.getLogger(__name__)


# TODO find a better way to get the namespace automatically?
_ns = '{http://Alma/XASDM/sdmbin}'


def basename_noext(path):
    return os.path.basename(os.path.splitext(path)[0])


def _stripns(tag):
    return re.sub('{.+}', '', tag)


def ant2bl(i, j=None):
    """Returns baseline index for given antenna pair.  Will accept
    two args, or a list/tuple/etc.  Uses 0-based indexing"""
    if j is None:
        (a1, a2) = sorted(i[:2])
    else:
        (a1, a2) = sorted((i, j))
    # could raise error if a2==a1, either are negative, etc
    return (a2*(a2-1))//2 + a1


def bl2ant(i):
    """Returns antenna pair for given baseline index.  All are 0-based."""
    a2 = int(0.5*(1.0+math.sqrt(1.0+8.0*i)))
    a1 = i - a2*(a2-1)//2
    return a1, a2


# This class provides access to EVLA/ALMA Binary Data Format (BDF)
# files.  The approach used here is based loosely on 'bdfparse.py'
# originally by Peter Williams.  So far this is somewhat targeted
# to EVLA/WIDAR data files, it may not handle all the ALMA variants.

class BDF(object):
    """
    Class representing a single BDF file.  For example:

        b = bdf.BDF('uid____evla_bdf_1433189755525')

    Individual integration data is returned as BDFIntegration objects via
    either b.get_integration(idx) or b[idx].  Other useful methods include:

        b.basebands      # list of baseband ids
        b.spws           # list of spectral windows per baseband
        b.numAntenna     # number of antennas
        b.numBaseline    # number of baselines
        b.numIntegration # number of integrations in file
        b.sdmDataHeader  # lxml objectify version of full header

    """

    def __init__(self, fname):
        self.fname = fname
        try:
            self.fp = open(fname, 'rb')
        except IOError:
            self.fp = None
        else:
            self.mmdata = mmap.mmap(self.fp.fileno(), 0, mmap.MAP_PRIVATE,
                                    mmap.PROT_READ)
            self.read_mime()
            self.parse_spws()

    # Size in bytes of each data element. In principle, the crossData
    # type needs to be read from the headers while all others have
    # pre-set values...
    bin_dtype_size = {
            'flags':           4, # INT32
            'actualTimes':     8, # INT64
            'actualDurations': 8, # INT64
            'zeroLags':        4, # FLOAT32
            'autoData':        4, # FLOAT32
            'crossData':       4, # FLOAT32 but should be determined from file
            }

    # Basic data type for each array.  Note, does not necessarily
    # match up with the sizes above due to how BDF defines the 'size'
    # attribute..
    bin_dtype = {
            'autoData':        numpy.float32,
            'crossData':       numpy.complex64,
            }

    @property
    def exists(self):
        return self.fp is not None

    def read_mime(self, full_read=False):
        if self.fp:
            self.fp.seek(0, 0)  # Go back to start
            if not self.fp.readline().decode('utf-8').startswith('MIME-Version:'):
                raise RuntimeError('Invalid BDF: missing MIME-Version')

            # First we need to read and parse only the main XML header in order
            # to get sizes of the binary parts.  Note, the info stored in
            # self.bin_size is in bytes, rather than the weird BDF units.
            mime_hdr = MIMEPart(self.fp).hdr
            self.top_mime_bound = mime_hdr.boundary
            sdmDataMime = MIMEPart(self.fp, boundary=self.top_mime_bound)
            if sdmDataMime.loc != 'sdmDataHeader.xml':
                raise RuntimeError('Invalid BDF: missing sdmDataHeader.xml')
            self.sdmDataHeader = objectify.fromstring(bytes(sdmDataMime.body, 'utf-8'))
            self.bin_size = {}
            self.bin_axes = {}
            for e in self.sdmDataHeader.iter():
                if 'size' in list(e.attrib.keys()) and 'axes' in list(e.attrib.keys()):
                    binname = _stripns(e.tag)
                    self.bin_size[binname] = int(e.attrib['size']) \
                        * self.bin_dtype_size[binname]
                    self.bin_axes[binname] = e.attrib['axes'].split()

            # For EVLA, we can read the first integration, note the file offset
            # in order to determine the size, then seek to each integration as
            # requested, rather than parsing the whole file here.
            if 'EVLA' in mime_hdr['Content-Description'][0] and not full_read:
                self.offset_ints = self.fp.tell()  # Offset in file to first integ
                self.mime_ints = [MIMEPart(self.fp,
                                           boundary=self.top_mime_bound,
                                           binary_size=self.bin_size,
                                           recurse=True), ]
            # Compute size of each integration section:
                self.size_ints = self.fp.tell() - self.offset_ints
                numints = int((os.path.getsize(self.fname)-self.offset_ints)//self.size_ints)
                self.mime_ints += [None, ]*(numints-1)

            # This is the more general way to do it that does not assume
            # each integration (including XML and MIME headers) has the
            # same size in the file.  In this case, go back to the beginning
            # and parse the whole MIME structure to map it out.
            else:
                self.fp.seek(0, 0)  # reset to start again
                full_mime = MIMEPart(self.fp,
                                     recurse=True, binary_size=self.bin_size)
                self.mime_ints = full_mime.body[1:]
        else:
            logger.warn('No BDF file found at {0}'.format(self.fname))


    def _raw(self, idx):
        if self.fp:
            if self.mime_ints[idx] is not None:
                    return self.mime_ints[idx]

            # Need to read this one
            self.fp.seek(self.offset_ints + idx*self.size_ints, 0)
            #self.mime_ints[idx] = self.read_mime_part(boundary=self.top_mime_bound,recurse=True)
            self.mime_ints[idx] = MIMEPart(self.fp,
                                           boundary=self.top_mime_bound,
                                           binary_size=self.bin_size,
                                           recurse=True)
            return self.mime_ints[idx]
        else:
            logger.warn('No BDF file found at {0}'.format(self.fname))
            return None

    @property
    def projectPath(self):
        return self.sdmDataHeader.attrib['projectPath']

    @property
    def numIntegration(self):
        return len(self.mime_ints)

    @property
    def numAntenna(self):
        return int(self.sdmDataHeader.numAntenna)

    @property
    def numBaseline(self):
        return (self.numAntenna*(self.numAntenna-1))//2

    @property
    def startTime(self):
        return float(self.sdmDataHeader.startTime)/86400.0e9

    def parse_spws(self):
        self.basebands = []
        self.spws = []
        # Offsets into the cross and data arrays where this spw
        # can be found.  This info could be reconstructed later
        # but seems convenient to do it here.
        cross_offset = 0
        auto_offset = 0
        self.spws = []
        for bb in self.sdmDataHeader.dataStruct.baseband:
            bbname = bb.attrib['name']
            self.basebands.append(bbname)
            for spw_elem in bb.spectralWindow:
                # Build a list of spectral windows for each baseband
                spw = BDFSpectralWindow(spw_elem, cross_offset, auto_offset)
                cross_offset += spw.dsize('cross')
                auto_offset += spw.dsize('auto')
                self.spws.append(spw)

    def get_integration(self, idx):
        return BDFIntegration(self, idx)

    def __getitem__(self, idx):
        return self.get_integration(idx)

    def zerofraction(self, spwidx='all', type='cross'):
        """
        Return zero fraction for the entire BDF.  This is done by loading
        each integration's data so may take a while.
        """

        tot = 0.0
        for i in self:
            tot += i.zerofraction(spwidx, type)
        return tot / self.numIntegration

    def get_data(self, spwidx='all', type='cross', scrunch=False,
                 fscrunch=False, frange=None, trange=None, bar=False):
        """Returns an array containing all integrations for the specified
        spw and data type.  Takes a number of options:

          trange: tuple giving range of integrations to return (default=all).
          scrunch: if True, all requested integrations will be time-averaged.
          fscrunch: if True, data will be averaged over the channel axis.
          frange: range of channels to average if using fscrunch.
          bar: if True and the progressbar package is available, display
            a progress bar as data are loaded.

        If spwidx=='all' and no averaging was requested, then the dimensions
        of the returned array are: (time, baseline/antenna, spw, bin, channel,
        polarization).  If a single spw is selected, the spw axis is omitted.
        If time and/or freq averaging is selected, the time and/or channel
        axes are omitted.
        """
        chidx = -2  # index of spectral channels
        # Figure out ranges:
        if trange is None:
            i0 = 0
            i1 = self.numIntegration
        else:
            i0 = trange[0]
            i1 = trange[1]
        # Read first integration to get shapes, etc
        nsubout = i1 - i0
        subdat = self.get_integration(i0).get_data(spwidx, type)
        if scrunch:
            dshape = subdat.shape
        else:
            dshape = (nsubout,) + subdat.shape
        if fscrunch:
            dshape = dshape[:chidx] + dshape[chidx+1:]
        result = numpy.zeros(dshape, dtype=subdat.dtype)
        if bar and ProgressBar is not None:
            b = ProgressBar()
        else:
            b = lambda x: x
        for i in b(list(range(i0, i1))):
            if fscrunch:
                if frange is None:
                    dat = self.get_integration(i).get_data(spwidx, type).mean(chidx)
                else:
                    dat = self.get_integration(i).get_data(spwidx,
                                                           type).take(list(range(*frange)),
                                                                      axis=chidx).mean(chidx)
            else:
                dat = self.get_integration(i).get_data(spwidx, type)
            if scrunch:
                result += dat
            else:
                result[i-i0] = dat
        if scrunch:
            result /= float(nsubout)
        return result


class BDFSpectralWindow(object):
    """Class that represents spectral window information present in BDF files,
    including storing appropriate offsets into the main data array.  Should be
    initialized from the spectralWindow XML element from the main BDF
    header.

    Alternatively, a BDFSpectralWindow can be generated directly without
    reference to an existing XML element.  In this case, the following
    arguments should be filled in appropriately (names are same as in XML):
        numBin
        numSpectralPoint
        sw
        swbb
    And the npol argument should be set to either 2 or 4; other values are
    not currently handled.

    An XML Element can be generated (eg for output) using the to_xml()
    method.
    """

    # Example spw element:
    # <spectralWindow sw="1" swbb="AC_8BIT" sdPolProducts="RR RL LL" crossPolProducts="RR RL LR LL" numSpectralPoint="64" numBin="40" scaleFactor="1.000000" sideband="NOSB"/>

    def __init__(self, spw_elem, cross_offset=None, auto_offset=None,
                 numBin=None, numSpectralPoint=None, sw=None, swbb=None,
                 npol=None):
        if spw_elem is not None:
            self._attrib = spw_elem.attrib
        else:
            # Fill _attrib based on extra input:
            self._attrib = {}
            self._attrib['numBin'] = '%d' % numBin
            self._attrib['numSpectralPoint'] = '%d' % numSpectralPoint
            self._attrib['sw'] = '%d' % sw
            self._attrib['swbb'] = str(swbb)
            if npol == 4:
                self._attrib['sdPolProducts'] = 'RR RL LL'
                self._attrib['crossPolProducts'] = 'RR RL LR LL'
            elif npol == 2:
                self._attrib['sdPolProducts'] = 'RR LL'
                self._attrib['crossPolProducts'] = 'RR LL'
            else:
                raise RuntimeError("Don't know how to handle npol=%d" % npol)
            # Boilerplate for VLA?
            self._attrib['scaleFactor'] = '1.000000'
            self._attrib['sideband'] = 'NOSB'

        self.cross_offset = cross_offset
        self.auto_offset = auto_offset

    def to_xml(self):
        # Property?
        # Note this returns a standalone xml Element, _not_ a reference to
        # the original document structure that this was derived from.
        result = etree.Element('spectralWindow')
        for k, v in list(self._attrib.items()):
            result.attrib[k] = v
        return result

    @property
    def numBin(self):
        return int(self._attrib['numBin'])

    @property
    def numSpectralPoint(self):
        return int(self._attrib['numSpectralPoint'])

    # These two are not listed in the BDF spec but appear in EVLA data:
    @property
    def sw(self):
        return int(self._attrib['sw'])

    @property
    def swbb(self):
        return self._attrib['swbb']

    # Should uniquely identify a sw?
    @property
    def name(self):
        return self.swbb + '-' + str(self.sw)

    def npol(self, type):
        """Return number of polarization array elements for the given data
        type (cross or auto)."""
        try:
            if type[0].lower() == 'c':
                return len(self._attrib['crossPolProducts'].split())
            elif type[0].lower() == 'a':
                # Good enough?
                l = len(self._attrib['sdPolProducts'].split())
                return 4 if l == 3 else l  # 3==4 in BDF math! :)
        except KeyError:
            return 0

    def dshape(self, type):
        """Return shape tuple of data array for this spectral window,
        in number of data elements (real for auto, complex for cross).
        """

        return (self.numBin, self.numSpectralPoint, self.npol(type))

    def dsize(self, type):
        """Return size of data array for this spectral window, in number of
        data elements (real for auto, complex for cross)."""
        return numpy.product(self.dshape(type))

    @staticmethod
    def dims_match(spwlist, type):
        """Given a list of BDFSpectralWindow objects, return true if all
        of them have consistent array dimensions."""
        if len(spwlist) == 1:
            return True
        for spw in spwlist[1:]:
            if spwlist[0].dshape(type) != spw.dshape(type):
                return False
        return True


class BDFIntegration(object):
    """
    Describes and holds data for a single intgration within a BDF file.
    This should be derived from an existing BDF object using
    get_integration() or indexing, ie:

        b = bdf.BDF('some_file')

        # Get the 5th integration, these two are equivalent:
        i = b.get_integration(5)
        i = b[5]

        # Read the cross-corr data array for spectral window 0
        dat = i.get_data(0)

    Other potentially useful info:

        i.basebands            # list of baseband IDs
        i.spws                 # dict of spws per baseband
        i.numAntenna           # obvious
        i.numBaseline          # "
        i.sdmDataSubsetHeader  # lxml objectify version of full sub-header

    """

    def __init__(self, bdf, idx):
        # Get the main header
        self.sdmDataSubsetHeader = objectify.fromstring(
                bytes(bdf._raw(idx).body[0].body, 'utf-8'))
        # Copy some info from the BDF headers
        self.basebands = bdf.basebands
        self.spws = bdf.spws
        self.bin_axes = bdf.bin_axes
        self.numAntenna = bdf.numAntenna
        self.numBaseline = bdf.numBaseline
        # Get the binary data
        self.data = {}
        for m in bdf._raw(idx).body[1:]:
            btype = basename_noext(m.loc)
            bsize = bdf.bin_size[btype]  # size of the binary blob in bytes
            baxes = self.bin_axes[btype]
            # Determine outer step size for the array, either baselines
            # antennas of baseline+antenna.  We can't apply the other
            # dimensions here because the number of elements can vary
            # per spw.
            if baxes[0] == 'BAL' and baxes[1] == 'ANT':
                shape = (self.numBaseline+self.numAntenna, -1)
            elif baxes[0] == 'BAL':
                shape = (self.numBaseline, -1)
            elif baxes[0] == 'ANT':
                shape = (self.numAntenna, -1)
            else:
                shape = (-1,)  # Don't know what to do, just leave flat array
            self.data[btype] = numpy.frombuffer(bdf.mmdata[m.body:m.body+bsize],
                                                dtype=bdf.bin_dtype[btype]).reshape(shape)

    @property
    def projectPath(self):
        return self.sdmDataSubsetHeader.attrib['projectPath']

    @property
    def time(self):
        return float(self.sdmDataSubsetHeader.schedulePeriodTime.time)/86400.0e9

    @property
    def interval(self):
        return float(self.sdmDataSubsetHeader.schedulePeriodTime.interval)*1e-9

    def get_data(self, spwidx='all', type='cross'):
        """
        Return the data array for the given subset.  Inputs are:

            spwidx:    spw index within file
            type:      'cross' or 'auto' (default 'cross')

        The returned array shape is (nBl/nAnt, nBin, nSpp, nPol).

        If spwidx is 'all' and the array dimensions of all spectral
        windows match, all will be returned in a single array.  In this case
        the returned dimensions will be (nBl/nAnt, nSpw, nBin, nSpp, nPol).
        """
        if type[0].lower() == 'c':
            loc = 'crossData'
        elif type[0].lower() == 'a':
            loc = 'autoData'
        else:
            raise RuntimeError('Unsupported data type')
        if spwidx == 'all':
            if not BDFSpectralWindow.dims_match(self.spws, type):
                raise RuntimeError('BDFIntegration: ' +
                                   'mixed array dimensions, spws must be ' +
                                   'retrieved indivdually')
            dshape = (-1, len(self.spws)) + self.spws[0].dshape(type)
            return self.data[loc].reshape(dshape)
        spw = self.spws[spwidx]
        if loc == 'crossData':
            offs = spw.cross_offset
        elif loc == 'autoData':
            offs = spw.auto_offset
        else:
            raise RuntimeError('Unsupported data type')
        dsize = spw.dsize(type)
        dshape = (-1,) + spw.dshape(type)
        return self.data[loc][:, offs:offs+dsize].reshape(dshape)

    def zerofraction(self, spwidx='all', type='cross'):
        """Returns the fraction of data points in the integration that
        are exactly zero (generally this means they have been flagged
        or otherwise not recorded by the online systems).

        Note that for WIDAR autocorrelation data, the default is to only
        record half the antennas so will typically have ~50% zeros
        according to this function."""

        if type[0].lower() == 'c':
            loc = 'crossData'
        elif type[0].lower() == 'a':
            loc = 'autoData'

        if spwidx == 'all':
            dtmp = self.data[loc].ravel()
        else:
            dtmp = self.get_data(spwidx=spwidx, type=type).ravel()
        return float(len(dtmp) - numpy.count_nonzero(dtmp))/float(len(dtmp))

# Notes for generating BDFs from scratch:
#
# Example data header:
#<sdmDataHeader xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xl="http://www.w3.org/1999/xlink" xmlns:xv="http://Alma/XVERSION" xmlns="http://Alma/XASDM/sdmbin" xsi:schemaLocation="http://Alma/XASDM/sdmbin http://almaobservatory.org/XML/XASDM/sdmbin/2/sdmDataObject.xsd" xv:schemaVersion="2" xv:revision="1.1.2.1" xv:release="ALMA-6_1_0-B" mainHeaderId="sdmDataHeader" byteOrder="Little_Endian" projectPath="0/1/1/">
#  <startTime>4922977435000000000</startTime>
#  <dataOID xl:type="locator" xl:href="uid:///evla/bdf/1416260627675" xl:title="EVLA WIDAR correlator visibility data"/>
#  <dimensionality axes="TIM">1</dimensionality>
#  <execBlock xl:href="uid:///evla/bdf/1416260627675" xl:type="simple"/>
#  <numAntenna>27</numAntenna>
#  <correlationMode>CROSS_AND_AUTO</correlationMode>
#  <spectralResolution>FULL_RESOLUTION</spectralResolution>
#  <processorType>CORRELATOR</processorType>
#  <dataStruct xsi:type="CrossAndAutoData" apc="AP_UNCORRECTED">
#    <baseband name="AC_8BIT">
#      <spectralWindow sw="1" swbb="AC_8BIT" sdPolProducts="RR RL LL" crossPolProducts="RR RL LR LL" numSpectralPoint="64" numBin="40" scaleFactor="1.000000" sideband="NOSB"/>
#    </baseband>
#    <flags size="59400" axes="BAL ANT BAB SPW BIN STO"/>
#    <actualTimes size="59400" axes="BAL ANT BAB SPW BIN STO"/>
#    <actualDurations size="59400" axes="BAL ANT BAB SPW BIN STO"/>
#    <crossData size="7188480" axes="BAL BAB SPW BIN SPP STO"/>
#    <autoData size="276480" axes="ANT BAB SPW BIN SPP STO" normalized="false"/>
#  </dataStruct>
#</sdmDataHeader>

# Build a sdmDataHeader from scratch
_nsmap_hdr = {
        'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xl': 'http://www.w3.org/1999/xlink',
        'xv': 'http://Alma/XVERSION',
        None: 'http://Alma/XASDM/sdmbin'
        }


def _sdmDataHeader(time, uid, num_antenna, spws, path='0/1/1', cross=True,
                   auto=True):
    """Generate a sdmDataHeader XML element from the specified parameters:
        time: start time in SDM format (MJD ns)
        uid: unique ID for the BDF
        num_antenna: number of antennas in the data
        spws: list of BDFSpectralWindow objects; order matters!
    """
    _E = objectify.ElementMaker(annotate=False, nsmap=_nsmap_hdr)
    xl_type = '{%s}type' % _nsmap_hdr['xl']
    xsi_type = '{%s}type' % _nsmap_hdr['xsi']
    xl_href = '{%s}href' % _nsmap_hdr['xl']
    xl_title = '{%s}title' % _nsmap_hdr['xl']
    xsi_schemalocation = '{%s}schemaLocation' % _nsmap_hdr['xsi']
    xv_schemaversion = '{%s}schemaVersion' % _nsmap_hdr['xv']
    xv_revision = '{%s}revision' % _nsmap_hdr['xv']
    xv_release = '{%s}release' % _nsmap_hdr['xv']
    if cross and auto:
        corr_mode = 'CROSS_AND_AUTO'
        data_type = 'CrossAndAutoData'
    elif cross:
        corr_mode = 'CROSS_ONLY'
        data_type = 'CrossData'
    elif auto:
        corr_mode = 'AUTO_ONLY'
        data_type = 'AutoData'
    else:
        raise RuntimeError('No data type specified (cross or auto).')
    result = _E.sdmDataHeader(
            _E.startTime(time),
            _E.dataOID({
                xl_type: 'locator',
                xl_href: uid,
                xl_title: "EVLA WIDAR correlator visibility data"
                }),
            _E.dimensionality(1, axes="TIM"),
            _E.execBlock({xl_href: uid, xl_type: "simple"}),
            _E.numAntenna(num_antenna),
            _E.correlationMode(corr_mode),
            _E.spectralResolution('FULL_RESOLUTION'),
            _E.processorType('CORRELATOR'),
            _E.dataStruct({xsi_type: data_type, 'apc': 'AP_UNCORRECTED'}),
            # sdmDataHeader attributes
            {xsi_schemalocation: 'http://Alma/XASDM/sdmbin http://almaobservatory.org/XML/XASDM/sdmbin/2/sdmDataObject.xsd',
                xv_schemaversion: '2',
                xv_revision: '1.1.2.1',
                xv_release: 'ALMA-6_1_0-B',
                'mainHeaderId': 'sdmDataHeader',
                'byteOrder': 'Little_Endian',
                'projectPath': path}
            )
    # Now add the spws.  We are assuming they are sorted in the right
    # order already...
    cur_bb = None
    auto_size = 0
    cross_size = 0
    for s in spws:
        if s.swbb != cur_bb:
            cur_bb = s.swbb
            bb = _E.baseband(name=cur_bb)
            result.dataStruct.append(bb)
        bb.append(s.to_xml())
        auto_size += s.dsize('auto')
        cross_size += s.dsize('cross')
    num_baseline = num_antenna * (num_antenna-1) // 2
    auto_size *= num_antenna
    cross_size *= 2.0 * num_baseline
    if cross:
        result.dataStruct.append(_E.crossData(size='%d' % cross_size,
                                              axes="BAL BAB SPW BIN SPP STO"))
    if auto:
        result.dataStruct.append(_E.autoData(size='%d' % auto_size,
                                             axes="ANT BAB SPW BIN SPP STO",
                                             normalized="false"))
    return result

# Example data sub header:
#<sdmDataSubsetHeader xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:xl="http://www.w3.org/1999/xlink" xmlns="http://Alma/XASDM/sdmbin" xsi:type="BinaryCrossAndAutoDataFXF" projectPath="0/1/1/1/">
#  <schedulePeriodTime>
#    <time>4922977454422548000</time>
#    <interval>14350776000</interval>
#  </schedulePeriodTime>
#  <dataStruct ref="sdmDataHeader"/>
#  <crossData xl:href="0/1/1/1/crossData.bin" type="FLOAT32_TYPE"/>
#  <autoData xl:href="0/1/1/1/autoData.bin"/>
#</sdmDataSubsetHeader>

# Build a sdmDataSubsetHeader element from scratch
_nsmap_subhdr = {
        'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xl': 'http://www.w3.org/1999/xlink',
        None: 'http://Alma/XASDM/sdmbin'
        }


def _sdmDataSubsetHeader(time, interval, cross=True, auto=True,
                         path='0/1/1/1/'):
    _E = objectify.ElementMaker(annotate=False, nsmap=_nsmap_subhdr)
    xl_href = '{%s}href' % _nsmap_subhdr['xl']
    xsi_type = '{%s}type' % _nsmap_subhdr['xsi']
    result = _E.sdmDataSubsetHeader(
            _E.schedulePeriodTime(
                _E.time(time),
                _E.interval(interval)
                ),
            _E.dataStruct(ref='sdmDataHeader'),
            projectPath=path)
    if cross:
        result.append(_E.crossData({
            'type': 'FLOAT32_TYPE',
            xl_href: path + 'crossData.bin'
            }))
        result.attrib[xsi_type] = 'BinaryCrossDataFXF'
    if auto:
        result.append(_E.autoData({
            xl_href: path + 'autoData.bin'
            }))
        result.attrib[xsi_type] = 'BinaryAutoDataFXF'
    if cross and auto:
        result.attrib[xsi_type] = 'BinaryCrossAndAutoDataFXF'
    return result


class BDFWriter(object):
    """
    Write a BDF file.
    """
    def __init__(self, path, fname=None, bdf=None, start_mjd=None, uid=None,
                 num_antenna=None, spws=None, scan_idx=None, subscan_idx=1,
                 corr_mode=None):
        """Init BDFWrite with output filename (fname).  If the bdf
        argument contains a BDF object, its header is copied for the
        output file.  Otherwise the following arguments need to be
        filled in for the header:

            start_mjd: Start time of the BDF in MJD
            uid: UID to put in the header (eg, uid:///evla/bdf/1484080742396)
            num_antenna: Number of antennas
            spws: List of BDFSpectralWindow objects in correct order
            scan_idx: index of this scan in the SDM
            subscan_idx: idx of the subscan in the SDM (default 1)
            corr_mode:  'ca' for cross and auto, 'c' for cross, 'a' for auto
        """
        if fname is not None:
            self.fname = os.path.join(path, fname)
        else:
            self.fname = os.path.join(path,
#                                      uid.translate(string.maketrans(':/','__')))
                                      uid.replace(':/', '__').replace('/', '_'))
        self.fp = None
        self.curidx = 1
        self.mb1 = "MIME_boundary-1"
        self.mb2 = "MIME_boundary-2"
        self.len0 = 0
        self.len1 = 0
        self.len2 = 0
        self.sdmDataHeader = None
        if bdf is not None:
            self.sdmDataHeader = deepcopy(bdf.sdmDataHeader)
        else:
            cross = 'c' in corr_mode
            auto = 'a' in corr_mode
            path = '0/%d/%d/' % (scan_idx, subscan_idx)
            self.sdmDataHeader = _sdmDataHeader(int(start_mjd*86400.0e9),
                                                uid, num_antenna, spws,
                                                path=path, cross=cross,
                                                auto=auto)

    def write_header(self):
        """Open output and write the current header contents."""
        self.fp = open(self.fname, 'wb')
        tophdr = MIMEHeader()
        tophdr['MIME-Version'] = ['1.0', ]
        tophdr['Content-Type'] = ['multipart/mixed', 'boundary='+self.mb1]
        tophdr['Content-Description'] = [
                'EVLA/CORRELATOR/WIDAR/FULL_RESOLUTION', ]
        # How do we generate a new unique name?
        nsxl = self.sdmDataHeader.nsmap['xl']
        uid = self.sdmDataHeader.dataOID.attrib['{%s}href' % nsxl][5:]
        tophdr['Content-Location'] = ['http://evla.nrao.edu/wcbe/XSDM' + uid, ]
        self.fp.write(tophdr.tostring() + '\n')

        self.fp.write('--' + self.mb1 + '\n')
        xhdr = MIMEHeader()
        xhdr['Content-Type'] = ['text/xml', 'charset=utf-8']
        xhdr['Content-Location'] = ['sdmDataHeader.xml', ]
        self.fp.write(xhdr.tostring() + '\n')
        self.fp.write(etree.tostring(self.sdmDataHeader,
                                     standalone=True, encoding='utf-8') + '\n')

    def write_integration(self, bdf_int=None, mjd=None, interval=None,
                          data=None):
        """
        Input is a BDFIntegration object (bdf_int).  The projectPath will
        be updated so that it is consistent for the file being written but
        otherwise no changes are made to the contents.

        Alternately, rather than a bdf_int, the remaining arguments can be
        filled in (for creating BDFs from scratch):

          mjd: the MJD of the midpoint of the integaration
          interval: the duration of the integration (sec)
          data: a dict whose entries are the numpy data arrays.  The
            keywords should be one or both of 'crossData' and 'autoData'
            depending on whether cross-correlations, auto-correlations, or
            both are present in the data set.
        """

        # NOTES for doing this:  bdf_int does not really need to be a
        # BDFIntegration.  It needs to act like it in the following ways:
        # 1. It needs to have a sdmDataSubsetHeader lxml Element object
        # representing the sub header.
        # 2. It needs to have a data attribute which is a dict of numpy
        # arrays containing the actual data to be written.  It

        tophdr = MIMEHeader()
        tophdr['Content-Type'] = ['multipart/related', 'boundary='+self.mb2]
        tophdr['Content-Description'] = ['data and metadata subset', ]

        ppidx = self.sdmDataHeader.attrib['projectPath'] + '%d/' % self.curidx

        hdr = MIMEHeader()
        hdr['Content-Type'] = ['text/xml', 'charset=utf-8']
        hdr['Content-Location'] = [ppidx + 'desc.xml']
        if self.len0 == 0:
            self.len0 = len(hdr.tostring()) + 12
        nxpad = self.len0 - len(hdr.tostring())
        if nxpad < 0:
            raise RuntimeError('nxpad(0)<0')
        hdr['X-pad'] = ['*'*nxpad, ]

        # Copy or generate XML sub-header
        if bdf_int is not None:
            subhdr = deepcopy(bdf_int.sdmDataSubsetHeader)
            data = bdf_int.data
        else:
            cross = 'crossData' in list(data.keys())
            auto = 'autoData' in list(data.keys())
            subhdr = _sdmDataSubsetHeader(int(mjd*86400e9), int(interval*1e9),
                                          cross=cross, auto=auto)

        subhdr.attrib['projectPath'] = ppidx
        nsxl = subhdr.nsmap['xl']
        dtypes = []
        mhdr = {}
        for dtype in ('crossData', 'autoData'):
            try:
                loc = ppidx + dtype + '.bin'
                getattr(subhdr, dtype).attrib['{%s}href' % nsxl] = loc
                dtypes += [dtype, ]
                mhdr[dtype] = MIMEHeader()
                mhdr[dtype]['Content-Type'] = ['application/octet-stream']
                mhdr[dtype]['Content-Location'] = [loc]
            except AttributeError:
                pass

        # Figure out how much X-pad to add
        subhdr_str = etree.tostring(subhdr, standalone=True, encoding='utf-8')
        if self.len1 == 0:
            self.len1 = len(subhdr_str) + len(mhdr[dtypes[0]].tostring()) + 50
        nxpad = self.len1 - (len(subhdr_str) + len(mhdr[dtypes[0]].tostring()))
        mhdr[dtypes[0]]['X-pad'] = ['*'*nxpad, ]

        # Assumes at most 2 data types.. TODO make more general?
        if len(dtypes) > 1:
            if self.len2 == 0:
                self.len2 = len(mhdr[dtypes[1]].tostring()) + 12
            nxpad = self.len2 - len(mhdr[dtypes[1]].tostring())
            mhdr[dtypes[1]]['X-pad'] = ['*'*nxpad, ]

        # TODO should check that data sizes match up with header info..

        # Now write it all out..
        self.fp.write('--' + self.mb1 + '\n')
        self.fp.write(tophdr.tostring()+'\n')

        # XML subheader part
        self.fp.write('--' + self.mb2 + '\n')
        self.fp.write(hdr.tostring() + '\n')
        self.fp.write(subhdr_str)

        # Data parts
        for dtype in dtypes:
            self.fp.write('\n--' + self.mb2 + '\n')
            self.fp.write(mhdr[dtype].tostring() + '\n')
            self.fp.write(data[dtype])

        # Close out mime
        self.fp.write('\n--' + self.mb2 + '--\n')
        self.curidx += 1

    def close(self):
        self.fp.write('--' + self.mb1 + '--\n')
        self.fp.close()



from __future__ import print_function, division, absolute_import, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input, int #, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import os
import warnings
import numpy as np

# TODO maybe reconsider this dependency, although the MJD class is convenient
from psrchive import MJD

# The big kludge-pile:
# _bin_file = "/lustre/aoc/sciops/pdemores/tests/phase_bins/binning_period.dat"
_bin_file = "/home/mchost/evla/scripts/opt/2016/09/16B-248/polyco/16B-248.period.log"
_bin_epochs = None   # list of MJDs
_bin_periods = None  # list of periods (s)
_mjd1970 = 40587
_clk_per_sec = int(64000000)
_clk_per_day = _clk_per_sec * int(86400)

# Routines to support pulsar binned/gated SDM data


def _get_epoch_period(mjd):
    """Get the closest binning reference epoch and period for a specified
    MJD.  This is kind of a hack until this info is stored properly in the
    SDM.  Returns a tuple (epoch, period, diff)"""
    global _bin_epochs, _bin_periods
    if _bin_epochs is None:
        _bin_epochs = []
        _bin_periods = []
        for l in open(_bin_file, 'rb'):
            # _bin_epochs += [MJD(l.split()[0]),]
            # _bin_periods += [float(l.split()[1]),]
            if l.startswith('epoch '):
                epoch_clk = int(l.split()[1])
                mjd_int = epoch_clk//_clk_per_day + _mjd1970
                mjd_frac = (epoch_clk % _clk_per_day)/float(_clk_per_day)
            elif l.startswith('target period '):
                p = float(l.split()[3])/float(_clk_per_sec)
                _bin_epochs += [MJD(mjd_int, mjd_frac), ]
                _bin_periods += [p, ]

    # Assumes input is a psrchive MJD.
    idx = np.argmin([abs((mjd-ep).in_seconds()) for ep in _bin_epochs])
    return (_bin_epochs[idx], _bin_periods[idx],
            (_bin_epochs[idx]-mjd).in_seconds())


# Kludge-pile version 2:
class BinLog(object):

    _mjd1970 = 40587
    _clk_per_sec = int(64000000)
    _clk_per_day = _clk_per_sec * int(86400)

    def __init__(self, sdmname,
                 logdir='/lustre/aoc/sciops/pdemores/binlogs'):
        self._logdir = logdir
        self.sdmname = sdmname
        # Following arrays contains info from each line of binlog file
        self.scanidx = []
        self.epoch = []
        self.period = []
        self._read()

    @classmethod
    def _clk_to_mjd(cls, clk):
        mjd_int = clk//cls._clk_per_day + cls._mjd1970
        mjd_frac = (clk % _clk_per_day)/float(_clk_per_day)
        return MJD(mjd_int, mjd_frac)

    def _read(self):
        # Actually read the file
        for l in open(os.path.join(self._logdir, self.sdmname+'.binlog'), 'rb'):
            (_sdmname, _idx, _epoch_clk, _period_us, _period_clk) = l.split()
            if _sdmname != self.sdmname:
                continue
            self.scanidx.append(_idx)
            self.epoch.append(self._clk_to_mjd(int(_epoch_clk)))
            self.period.append(float(_period_clk)/float(self._clk_per_sec))

    def epoch_period(self, mjd):
        idx = np.argmin([abs((mjd-ep).in_seconds()) for ep in self.epoch])
        tdiff = (self.epoch[idx] - mjd).in_seconds()
        return self.epoch[idx], self.period[idx], tdiff


def dm_delay(dm, freq1, freq2=np.inf):
    """Return dispersion delay in seconds from freq2 to freq1 for given
    DM.  Freqs in MHz, can be inf."""
    return (dm/0.000241)*(1.0/(freq1*freq1) - 1.0/(freq2*freq2))


def rotate_phase(data, turns, axis=1, method='lin'):
    if method == 'lin':
        return rotate_phase_lin(data, turns, axis=axis)
    elif method == 'fft':
        return rotate_phase_fft(data, turns, axis=axis)


def rotate_phase_lin(data, turns, axis=1):
    """
    Rotate the data array a given number of turns, assuming the specified
    axis is pulse phase and corresponds to a full turn of phase.  Currently
    uses linear interpolation, which may be a better choice than FFT-based
    rotation for poorly resolved features.
    """

    nbin = data.shape[axis]
    r = (turns - np.floor(turns)) * float(nbin)
    ri = int(r)
    rf = r - ri
    return (1.0-rf)*np.roll(data, ri, axis=axis) + rf*np.roll(data, ri+1,
                                                              axis=axis)


def rotate_phase_fft(data, turns, axis=1):
    """
    Rotate the data array a given number of turns, assuming the specified
    axis is pulse phase and corresponds to a full turn of phase.  Uses
    FFT-based rotation.
    """

    nbin = data.shape[axis]
    shape = [1, ] * len(data.shape)
    shape[axis] = nbin
    ff = np.arange(float(nbin))
    ff[np.where(ff > nbin/2.0)] -= float(nbin)
    phs = np.exp(-2.0j*np.pi*turns*ff).reshape(shape)
    fdata = np.fft.fft(data, axis=axis)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", np.ComplexWarning)
        return np.fft.ifft(fdata*phs, axis=axis).astype(data.dtype)


def dedisperse_array(data, dm, freq, period, bin_axis=1, freq_axis=2,
                     spw_axis=None):
    """
    Dedisperse a generic array of data, of which one axis represents an
    entire turn of pulse phase, even sampled into bins.  freq should be an
    array giving the frequencies in MHz.  Up to two separate freq axes are
    allowed, given by the spw_axis and freq_axis arguments.  If spw_axis is
    None (not used), freq should be a 1-D array of freqs.  If spw_axis is
    set, freq should have dims (nspw, nchan).
    """

    dp = -dm_delay(dm, freq)/period
    nchan = data.shape[freq_axis]
    fslice = [slice(None), ] * len(data.shape)
    fshape = list(data.shape)
    fshape[freq_axis] = 1
    if spw_axis is not None:
        nspw = data.shape[spw_axis]
        fshape[spw_axis] = 1
    else:
        nspw = 1
        dp = dp.reshape((1, -1))
    for ispw in range(nspw):
        if spw_axis is not None:
            fslice[spw_axis] = ispw
            dtmp0 = data.take([ispw, ], axis=spw_axis)
        for ichan in range(nchan):
            fslice[freq_axis] = ichan
            if spw_axis is None:
                dtmp = data.take(ichan, axis=freq_axis).reshape(fshape)
            else:
                dtmp = dtmp0.take([ichan, ], axis=freq_axis).reshape(fshape)
            data[fslice] = rotate_phase(dtmp, dp[ispw, ichan],
                                        axis=bin_axis).squeeze()

# def dedisperse(scan,dm,period=None,bar=lambda x: x):
#    """Dedisperse a scan at the given DM.  If period is not specified,
#    it will be looked up in the period vs time table (not implemented)."""
#    pass
#
# def rephase(scan,parfile):
#    """Rephase a scan to a new ephemeris given in parfile."""
#    pass



#! /usr/bin/env python
import sys, math
import numpy as np
import logging
import sdmpy
import sdmpy.pulsar
import psrchive

import argparse
par = argparse.ArgumentParser(
        description="Convert phase-binned SDM/BDF data to fold-mode PSRFITS.")
par.add_argument("sdmname", help="SDM to process")
#par.add_argument("-v", "--verbose", dest='loglevel',
#        action='store_const', const=logging.DEBUG,
#        default=logging.INFO,
#        help='Enable DEBUG output')
par.add_argument("-q", "--quiet", dest='loglevel',
        action='store_const', const=logging.WARN,
        default=logging.INFO,
        help='No INFO messages')
par.add_argument("-d", "--dm", type=float, default=0.0,
        help="dispersion measure (pc cm^-3) [%(default)s]")
par.add_argument("-p", "--period", type=float, default=0.0,
        help="period data were folded (s) [auto]")
par.add_argument("-P", "--polycos", action="store_true",
        help="use polyco file to adjust timestamps [false]")
args = par.parse_args()

logging.basicConfig(format="%(asctime)-15s %(levelname)8s %(message)s", 
        level=args.loglevel)

sdmname = args.sdmname.rstrip('/')
sdm = sdmpy.SDM(sdmname)
try:
    binlog = sdmpy.pulsar.BinLog(sdmname)
except IOError:
    binlog = None

# TODO update to read 4-pol data
npol = 2

def unroll_chans(data):
    """Unroll the spw and channel dims into a single frequency axis.
    Assumes incoming dimensions are (bl, spw, bin, chan, pol)"""
    data = data.transpose((0,2,1,3,4))
    data = data.reshape((data.shape[0], 
        data.shape[1], 
        data.shape[2]*data.shape[3],
        data.shape[4]))
    return data

for scan in sdm.scans():
    if 'CALIBRATE_PHASE' in scan.intents:
        logging.info('processing cal scan %s' % scan.idx)
        # dims (bl,spw,bin,chan,pol)
        dcal = scan.bdf.get_data(scrunch=True)[...,[0,-1]].mean(2,
                keepdims=True)
        dcal = unroll_chans(dcal)
        gcal = sdmpy.calib.gaincal(dcal,axis=0,ref=1)
    elif 'OBSERVE_TARGET' in scan.intents:
        logging.info('processing target scan %s' % scan.idx)

        psr = str(scan.source)

        # Assumes all spws have same number of bins
        nbin = int(scan.bdf.spws[0].numBin)

        # Assumes all spws have same number of chans
        freqs = scan.freqs().ravel()/1e6
        chanidx = np.argsort(freqs)
        nchan = freqs.size
        bw = sum([scan.spw(i).totBandwidth/1e6 for i in range(len(scan.spws))])

        # Initialize archive output
        arch = psrchive.Archive_new_Archive("ASP")
        arch.resize(0,npol,nchan,nbin)
        arch.set_source(psr)
        arch.set_dispersion_measure(args.dm)
        arch.set_coordinates(psrchive.sky_coord(*scan.coordinates))
        arch.set_centre_frequency(0.5*(freqs.max() + freqs.min()))
        arch.set_bandwidth(bw)
        arch.set_telescope('vla')
        arch.set_state('PPQQ')

        iout = 0

        # Try to load polycos
        try:
            polycofile = '%s/%s.%d.polyco' % (binlog._logdir,
                    sdmname, int(scan.idx))
            polys = psrchive.polyco(polycofile)
            logging.info('Read polycos from %s' % polycofile)
        except Exception as ex:
            logging.info("Couldn't load polycos: " + repr(ex))
            polys = None

        bdf = scan.bdf
        arch.resize(arch.get_nsubint() + bdf.numIntegration - 1)
        for isub in range(1,bdf.numIntegration):
            bdfsub = bdf[isub]
            dpsr = bdfsub.get_data()[...,[0,-1]]
            dpsr = unroll_chans(dpsr)
            sdmpy.calib.applycal(dpsr,gcal,phaseonly=True)
            dpsr = np.ma.masked_array(dpsr,dpsr==0.0)
            mpsr = np.real(dpsr.mean(0))
            subint = arch.get_Integration(iout)
            epoch_bdf = psrchive.MJD(float(bdfsub.time)) # only approx
            if args.period>0.0:
                #epoch = psrchive.MJD(epoch_bdf.intday())
                epoch = epoch_bdf
                p = args.period
                dt = 0.0
                logging.info('Using constant period')
            else:
                if binlog is not None:
                    # Apply polyco time adjust
                    if args.polycos and polys is not None:
                        dt = (polys.period(epoch_bdf) 
                                * polys.phase(epoch_bdf).fracturns())
                        epoch = epoch_bdf - dt
                        p = 0.0
                    else:
                        (epoch,p,dt) = binlog.epoch_period(epoch_bdf)
                else:
                    (epoch,p,dt) = sdmpy.pulsar._get_epoch_period(epoch_bdf)
                logging.info('Using epoch/period from dt=%.3fs' % dt)

            # These were used for testing dedispersion:
            #sdmpy.pulsar._dedisperse_array(mpsr, args.dm, freqs, p,
            #        bin_axis=0,freq_axis=1)
            #sdmpy.pulsar._dedisperse_array(mpsr, args.dm, freqs_spw, p,
            #        bin_axis=2,freq_axis=3,spw_axis=1)
            #mpsr = unroll_chans(mpsr)[0,...]

            subint.set_epoch(epoch)
            subint.set_duration(bdfsub.interval)
            if p>0.0: subint.set_folding_period(p)
            if isub==0:
                wt=0.0
            else:
                wt=1.0
            for ochan in range(nchan):
                ichan = chanidx[ochan]
                subint.set_centre_frequency(ochan,freqs[ichan])
                for ipol in range(npol):
                    prof = subint.get_Profile(ipol,ochan)
                    prof.get_amps()[:] = mpsr[:,ichan,ipol]
                subint.set_weight(ochan,wt)
            iout += 1

        if polys is not None:
            arch.set_model(polys,False)
        outputname = sdmname + '.%03d.fits' % int(scan.idx)
        logging.info("unloading '%s'" % outputname)
        arch.unload(outputname)



#! /usr/bin/env python

# Dedisperse and select individual integrations from a SDM 
# data set.
# PBD 09/2016

import os, sys, shutil
import argparse
import numpy as np
import sdmpy

par = argparse.ArgumentParser()
par.add_argument("sdmname", help="SDM to process")
par.add_argument("-i", "--idx", help="scan:int(:length) to store", 
        action="append", default=[])
par.add_argument("-e", "--ext", 
        help="extension to add for output [%(default)s]",
        default="ddcut")
par.add_argument("-d", "--dm", type=float, default=556.9,
        help="dispersion measure [%(default)s]")
par.add_argument("-t", "--time", type=float, default=0.0,
        help="keep all data with integration > time")
args = par.parse_args()

sdmname = args.sdmname.rstrip('/')

sdm = sdmpy.SDM(sdmname)

# dict of scan/integrations to keep
#keep = {'7': 9514,}
keep = {}
keeplen = {}
for i in args.idx:
    stuff = i.split(':')
    scan = stuff[0]
    iidx = stuff[1]
    try:
        length = stuff[2]
    except IndexError:
        length = 1
    keep[scan] = int(iidx)
    keeplen[scan] = int(length)

def dm_delay(dm,freq1,freq2=np.inf):
    """Return dispersion delay in seconds from freq2 to freq1 for given
    DM.  Freqs in MHz, can be inf."""
    return (dm/0.000241)*(1.0/(freq1*freq1) - 1.0/(freq2*freq2))

dm = args.dm

sdmout = sdmname + '.' + args.ext
bdfoutpath = sdmout + '/ASDMBinary'

os.mkdir(sdmout)
os.mkdir(bdfoutpath)

for scan in sdm.scans():
    print "Processing '%s' scan %s:" % (sdmname, scan.idx)
    try:
        bdf = scan.bdf
    except IOError:
        print "Error reading bdf for scan %s, skipping" % (scan.idx,)
        continue

    bdfoutname = bdfoutpath + '/' + os.path.basename(scan.bdf_fname)

    # Check either intent or int time, copy full scan if matching
    if (args.time>0 and bdf[0].interval>args.time):
        print "  copying scan %s" % (scan.idx,)
        shutil.copy(scan.bdf_fname, bdfoutname)
        continue

    if scan.idx not in keep.keys():
        print "  no integrations to keep for scan %s, skipping" % (scan.idx,)
        # Fill in X1
        scan._main.dataUID.EntityRef.attrib['entityId'] = 'uid://evla/bdf/X1'
        continue

    # If we got here, need to select appropriate integration(s), dedisperse
    # and output.

    dt = bdf[0].interval # in sec
    freqs = scan.freqs()/1e6
    delays = dm_delay(dm,freqs,freqs.max())
    delays_samp = np.rint(delays/dt)
    uniq_delays_samp = sorted(set(delays_samp.ravel()))
    nint_read = delays_samp.max()+1 # number of integrations to read per event

    nout = keeplen[scan.idx]
    for offs in range(nout):

        int0 = keep[scan.idx] + offs
        bdfint = bdf[int0]

        int_time = bdfint.sdmDataSubsetHeader.schedulePeriodTime.time
        int_interval = bdfint.sdmDataSubsetHeader.schedulePeriodTime.interval

        if offs==0:
            t0 = int_time - int_interval/2
            bdfout = sdmpy.bdf.BDFWriter(bdfoutname, bdf=bdf)
            bdfout.sdmDataHeader.startTime = t0
            bdfout.write_header()

        t1 = int_time + int_interval/2

        for dtype in bdfint.data.keys():
            # Copies the zero-delay data array
            bdfint.data[dtype] = bdfint.data[dtype].copy()
            dat0 = bdfint.get_data(type=dtype)
            # loop over the delays, copy data
            for ii in uniq_delays_samp:
                if (int0+ii)>bdf.numIntegration: continue
                dat = bdf[int(int0+ii)].get_data(type=dtype)
                fidx = np.where(delays_samp == ii)
                # May be a more clever numpy-ish way to do this?
                for jj in range(len(fidx[0])):
                    dat0[:,fidx[0][jj],0,fidx[1][jj],:] = \
                            dat[:,fidx[0][jj],0,fidx[1][jj],:]

        # Write this integration
        bdfout.write_integration(bdfint)

    bdfout.close()

    # update SDM entries with corect number of integrations
    scan._main.numIntegration = nout
    scan._main.dataSize = os.path.getsize(bdfoutname)
    scan._subscan.numIntegration = nout
    scan._subscan.numSubintegration = ('1 %d' % nout) + ' 0'*nout

    # Why must time be stored in four separate places...??
    scan._main.time = t0 # correct?
    scan._main.interval = int_interval*nout
    scan._scan.startTime = t0
    scan._scan.endTime = t1
    scan._subscan.startTime = t0
    scan._subscan.endTime = t1

# Write out new SDM tables
sdm.write(sdmout)

